from huggingface_hub import snapshot_download
from peft import PeftModel, LoraConfig, get_peft_model, prepare_model_for_kbit_training
from transformers import AutoModelForCausalLM, AutoTokenizer, AutoConfig, BitsAndBytesConfig
import torch
import bitsandbytes as bnb
from collections import defaultdict
from modelscope import snapshot_download as snapshot_download_ms
from lazydl.utils.log import Logger
from lazydl.utils.result import Result
from typing import Union

logger = Logger(__name__)


def download_model(model_name, cache_dir=None, endpoint='https://hf-mirror.com', use_modelscope=False):
    """下载 HuggingFace 模型权重，支持断点续传，适用于无法连接 HuggingFace 服务器的情况
 
    Args:
        model_name (_type_): 模型名称
        cache_dir (_type_, optional): 缓存地址，默认为 ~/.cache/huggingface. Defaults to None.
        endpoint (str, optional): 下载节点地址. Defaults to 'https://hf-mirror.com'.
        use_modelscope: 是否使用 ModelScope 下载

    Returns:
        _type_: _description_
    """
    if use_modelscope:
        return snapshot_download_ms(model_id=model_name,
                            ignore_file_pattern=["*.msgpack", "*.h5", "*.ot"],
                            cache_dir=cache_dir)
        
    return snapshot_download(repo_id=model_name,
                            repo_type='model',
                            ignore_patterns=["*.msgpack", "*.h5", "*.ot"],
                            local_dir=cache_dir,
                            local_dir_use_symlinks=False,  # 不转为缓存乱码的形式, auto, Small files (<5MB) are duplicated in `local_dir` while a symlink is created for bigger files.
                            endpoint=endpoint,
                            etag_timeout=60,
                            resume_download=True)
    
    
    
    
def merge_lora_to_base_model(model_name_or_path, adapter_name_or_path, save_path):
    """使用该脚本，将lora的权重合并大base model中

    Args:
        model_name_or_path (_type_): base model 的名称或路径
        adapter_name_or_path (_type_): lora 的名称或路径
        save_path (_type_): 合并后的模型保存路径
    """
    config = AutoConfig.from_pretrained(model_name_or_path, trust_remote_code=True)
    tokenizer = AutoTokenizer.from_pretrained(
        model_name_or_path,
        trust_remote_code=True,
        # llama不支持fast
        use_fast=False if config.model_type == 'llama' else True
    )
    model = AutoModelForCausalLM.from_pretrained(
        model_name_or_path,
        trust_remote_code=True,
        low_cpu_mem_usage=True,
        torch_dtype=torch.float16,
        device_map='auto',
        # device_map={'': 'cpu'}
    )
    model = PeftModel.from_pretrained(model, adapter_name_or_path, device_map={'': 'cpu'})
    model = model.merge_and_unload()

    tokenizer.save_pretrained(save_path)
    model.save_pretrained(save_path)
    logger.info(f"\nMerged model and tokenizer have saved to {save_path}")
    return model
    
    
def load_model_and_tokenizer(model_name_or_path, load_in_4bit=False, 
                             adapter_name_or_path=None, use_qlora=False,
                             gradient_checkpointing=True, lora_rank=64, lora_alpha=16, 
                             lora_dropout=0.05, device_map='auto'):
    """加载模型，支持4bit

    Args:
        model_name_or_path (_type_): 模型名称或路径
        load_in_4bit (bool, optional): 是否使用 4bit 量化进行推理. Defaults to False.
        adapter_name_or_path (_type_, optional):  Lora 权重路径. Defaults to None.
        lora_rank：qlora矩阵的秩。一般设置为8、16、32、64等，在qlora论文中作者设为64。越大则参与训练的参数量越大，一般来说效果会更好，但需要更多显存，。
        lora_alpha: qlora中的缩放参数。一般设为16、32即可。
        lora_dropout: lora权重的dropout rate。
        learning_rate：qlora中的学习率设置更大一些，一般为1e-4、2e-4。
        gradient_checkpointing：如果显存捉襟见肘，可以开启。以时间换空间，模型不缓存激活状态，会进行两次forward计算，以节省显存。

    Returns:
        Model: 模型
        Tokenizer: 分词器
    """
    logger.info("加载模型....")
    if load_in_4bit or use_qlora:
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_compute_dtype=torch.float16,
            bnb_4bit_use_double_quant=True,
            bnb_4bit_quant_type="nf4",
            llm_int8_threshold=6.0,
            llm_int8_has_fp16_weight=False,
        )
    else:
        quantization_config = None

    # 加载base model
    model = AutoModelForCausalLM.from_pretrained(
        model_name_or_path,
        load_in_4bit=load_in_4bit,
        trust_remote_code=True,
        low_cpu_mem_usage=True,
        torch_dtype=torch.float16,
        device_map=device_map,
        quantization_config=quantization_config
    )

    # 加载adapter
    if not use_qlora and adapter_name_or_path is not None:
        model = PeftModel.from_pretrained(model, adapter_name_or_path)
        
    if use_qlora:
        # casts all the non int8 modules to full precision (fp32) for stability
        model = prepare_model_for_kbit_training(model, use_gradient_checkpointing=gradient_checkpointing)
        print(f'memory footprint of model: {model.get_memory_footprint()/(1024*1024*1024)} GB')
        # 找到所有需要插入adapter的全连接层
        target_modules = find_all_linear_names(model)
        # 初始化lora配置
        config = LoraConfig(
            r=lora_rank,
            lora_alpha=lora_alpha,
            target_modules=target_modules,
            lora_dropout=lora_dropout,
            bias="none",
            task_type="CAUSAL_LM",
        )
        model = get_peft_model(model, config)
        # model.print_trainable_parameters()
        model.config.torch_dtype = torch.float32
        
    # 加载tokenzier
    tokenizer = AutoTokenizer.from_pretrained(
        model_name_or_path,
        trust_remote_code=True,
        # llama不支持fast
        use_fast=False if model.config.model_type == 'llama' else True
    )
    # QWenTokenizer比较特殊，pad_token_id、bos_token_id、eos_token_id均为None。eod_id对应的token为<|endoftext|>
    if tokenizer.__class__.__name__ == 'QWenTokenizer':
        tokenizer.pad_token_id = tokenizer.eod_id
        tokenizer.bos_token_id = tokenizer.eod_id
        tokenizer.eos_token_id = tokenizer.eod_id
    elif tokenizer.__class__.__name__ == 'LlamaTokenizer':
        assert tokenizer.eos_token_id is not None
        assert tokenizer.bos_token_id is not None
        tokenizer.pad_token_id = tokenizer.unk_token_id if tokenizer.pad_token_id is None else tokenizer.pad_token_id
    # ChatGLMTokenizer不需要设置，仅设置其他tokenizer
    elif tokenizer.__class__.__name__ != 'ChatGLMTokenizer':
        assert tokenizer.eos_token_id is not None
        assert tokenizer.bos_token_id is not None
        tokenizer.pad_token_id = tokenizer.eos_token_id if tokenizer.pad_token_id is None else tokenizer.pad_token_id
        
    # # 部分tokenizer没有pad_token_id
    # if tokenizer.pad_token_id is None:
    #     tokenizer.pad_token_id = tokenizer.unk_token_id
    # # 部分tokenizer的pad_token_id与eos_token_id相同，如InternLM，会导致无法计算eos_token_id的loss。将pad_token_id设为unk_token_id
    # if tokenizer.pad_token_id == tokenizer.eos_token_id and tokenizer.unk_token_id is not None:
    #     tokenizer.pad_token_id = tokenizer.unk_token_id
    # # 如果两者相同，模型训练时不会计算eos_token_id的loss
    # if tokenizer.pad_token_id == tokenizer.eos_token_id:
    #     raise Exception('pad_token_id should not be equal to eos_token_id')

    logger.info("模型准备就绪")
    return model, tokenizer


def verify_model_dtype(model):
    """
    查看模型种各种类型的参数的情况
    """
    dtype2param_num = defaultdict(int)  # 每种数据类型的参数量
    dtype2param_name = defaultdict(list)  # 每种数据类型的参数名称
    dtype2trainable_param_num = defaultdict(int)  # 每种数据类型参与训练的参数量
    dtype2trainable_param_name = defaultdict(list)  # 每种数据类型参与训练的参数名称
    for name, p in model.named_parameters():
        dtype = p.dtype
        dtype2param_num[dtype] += p.numel()
        dtype2param_name[dtype].append(name)
        if p.requires_grad:
            dtype2trainable_param_num[dtype] += p.numel()
            dtype2trainable_param_name[dtype].append(name)
    # 统计全部参数中，各种类型参数分布
    total = 0
    print('verify all params of the model')
    for k, v in dtype2param_num.items():
        total += v
    for k, v in dtype2param_num.items():
        print(k, v, v / total)
    for k, v in dtype2trainable_param_name.items():
        print(k, v)

    print()
    # 统计可训练参数中，各种类型参数分布
    print('verify trainable params the model')
    total_trainable = 0
    for k, v in dtype2trainable_param_num.items():
        total_trainable += v
    for k, v in dtype2trainable_param_num.items():
        print(k, v, v / total_trainable)
    for k, v in dtype2trainable_param_num.items():
        print(k, v)
        

def find_all_linear_names(model):
    """
    找出所有全连接层，为所有全连接添加adapter
    """
    cls = bnb.nn.Linear4bit
    lora_module_names = set()
    for name, module in model.named_modules():
        if isinstance(module, cls):
            names = name.split('.')
            lora_module_names.add(names[0] if len(names) == 1 else names[-1])

    if 'lm_head' in lora_module_names:  # needed for 16-bit
        lora_module_names.remove('lm_head')
    return list(lora_module_names)



class Pipeline:
    def __init__(self, 
                 model: object = None,
                 tokenizer: object = None,
                 model_name_or_path: str = None,
                 use_qlora: bool = True,
                 load_in_4bit: bool = False,
                 device_map: str = "auto",
                 adapter_name_or_path: str = None,
                 gradient_checkpointing: bool = True, 
                 lora_rank: int = 64, 
                 lora_alpha: int = 16, 
                 lora_dropout: int = 0.05, 
                 *args, **kargs):
        if model is not None and tokenizer is not None:
            self.model, self.tokenizer = model, tokenizer
        else:
            self.model, self.tokenizer = load_model_and_tokenizer(model_name_or_path, 
                                                                    use_qlora=use_qlora, load_in_4bit=load_in_4bit, device_map=device_map,
                                                                    gradient_checkpointing=gradient_checkpointing,
                                                                    adapter_name_or_path=adapter_name_or_path,
                                                                    lora_rank=lora_rank, 
                                                                    lora_alpha=lora_alpha, 
                                                                    lora_dropout=lora_dropout)
        self.model = self.model.eval()
        
        
    def generate(self, 
                 user_inputs: Union[str, list] = None, 
                 input_ids: list[list] = None,
                 max_length: int =512, 
                 top_k: int = 10, 
                 top_p: int = 0.9, 
                 temperature: int = 0.7, 
                 num_beams: int = 1,
                 num_return_sequences: int = 1,
                 *args, **kargs):
        """
        Generate text using the model
        """
        if input_ids is None:
            user_inputs_type = type(user_inputs)
            if user_inputs_type == str:
                user_inputs = [user_inputs]
            input_ids = self.tokenizer.batch_encode_plus(user_inputs, max_length=max_length, truncation=True, add_special_tokens=True)["input_ids"]

        else:
            user_inputs_type = type(input_ids[0])
            
            
        # 找出batch中的最大长度
        lengths = [len(x) for x in input_ids]
        # 取出batch中的最大长度
        batch_max_len = min(max(lengths), max_length)


        for input_index, ids in enumerate(input_ids):
            input_id_pad_len = batch_max_len - len(ids)
            input_ids[input_index] = [self.tokenizer.pad_token_id] * input_id_pad_len + ids
        
        input_ids = torch.tensor(input_ids, dtype=torch.long)
        input_ids = input_ids.to(self.model.device)
        
        outputs = self.model.generate(input_ids,
                                        max_new_tokens=max_length,
                                        do_sample=True,
                                        top_k=top_k,
                                        top_p=top_p,
                                        temperature=temperature,
                                        num_beams=num_beams,
                                        num_return_sequences=num_return_sequences,
        )
        model_input_ids_len = input_ids.size(1)
        response_ids = outputs[:, model_input_ids_len:]
        model_responses = self.tokenizer.batch_decode(response_ids, skip_special_tokens=True)
        if user_inputs_type == str:
            model_responses = model_responses[0]
        return model_responses
    


from lazydl.utils.result import Result
from lazydl.utils.log import Logger
from datasets import Dataset
import pandas as pd
from statistics import mean
import re
import string
from collections import Counter
from bert_score import score
from nltk import word_tokenize
from nltk.translate.bleu_score import corpus_bleu, sentence_bleu, SmoothingFunction
from nltk.translate.meteor_score import meteor_score as meteor_scorer
from nltk.util import ngrams
from rouge import Rouge
from sacrebleu.metrics import BLEU, CHRF
import sacrebleu
import spacy
from rich.console import Console
from rich.table import Table
from typing import Union

log = Logger(__name__) 


class DistCritic:
    def __init__(self) -> None:
        self.name = "dist"
        
    def score_step(self, candidates: list[str], eval_result: Result, n_gram: int = 2, *args, **kwargs) -> Result:
        ngram_freqs = {}  
        ngram_len = 0  
        for candidate in candidates:
            for ngram in ngrams(word_tokenize(candidate), n_gram):
                ngram_freqs[ngram] = ngram_freqs.get(ngram, 0) + 1
                ngram_len += 1
        # number of unique ngrams
        uniq_ngrams = len([val for val in ngram_freqs.values() if val == 1])
        distinct_ngram = len(ngram_freqs) / ngram_len if ngram_len > 0 else 0
        
        dist_score = round(distinct_ngram, 4)
        eval_result.merge({
            "DIST" + str(n_gram): dist_score
        })
        return eval_result

    def score(self, candidates: list[str], references: list[str], eval_result: Result, n_gram: int = 2, *args, **kwargs) -> Result:
        for step in range(n_gram):
            eval_result = self.score_step(candidates, eval_result, n_gram=step+1)
        
        return eval_result


class CorpusBleuCritic:
    def __init__(self) -> None:
        self.name = "corpus_bleu"

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        bleu1 = 0.0
        bleu2 = 0.0
        bleu3 = 0.0
        bleu4 = 0.0
        ref_list, dec_list = [], []
        for i in range(len(candidates)):
            dec_list.append(word_tokenize(candidates[i]))
            if type(references[i]) is list:
                tmp = []
                for ref in references[i]:
                    tmp.append(word_tokenize(ref))
                ref_list.append(tmp)
            else:
                ref_list.append([word_tokenize(references[i])])
        bleu1 = round(corpus_bleu(ref_list, dec_list, weights=(1, 0, 0, 0)) * 100, 4)
        bleu2 = round(corpus_bleu(ref_list, dec_list, weights=(0, 1, 0, 0)) * 100, 4)
        bleu3 = round(corpus_bleu(ref_list, dec_list, weights=(0, 0, 1, 0)) * 100, 4)
        bleu4 = round(corpus_bleu(ref_list, dec_list, weights=(0, 0, 0, 1)) * 100, 4)
        mean_corpus_bleu = round((bleu1 + bleu2 + bleu3 + bleu4) / 4, 4)
        
        eval_result.add(
            corpus_bleu1=bleu1,
            corpus_bleu2=bleu2,
            corpus_bleu3=bleu3,
            corpus_bleu4=bleu4,
            mean_corpus_bleu=mean_corpus_bleu,
        )
        
        return eval_result
    


class SentenceBleuCritic:
    def __init__(self) -> None:
        self.name = "sentence_bleu"

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        bleu1, bleu2, bleu3, bleu4 = 0.0, 0.0, 0.0, 0.0
        ref_list, dec_list = [], []
        for i in range(len(candidates)):
            dec_list.append(word_tokenize(candidates[i]))
            if type(references[i]) is list:
                tmp = []
                for ref in references[i]:
                    tmp.append(word_tokenize(ref))
                ref_list.append(tmp)
            else:
                ref_list.append([word_tokenize(references[i])])

        for example_id, (label, pred) in enumerate(zip(ref_list, dec_list)):
            bleu1 += sentence_bleu(
                label,
                pred,
                weights=[1, 0, 0, 0],
                smoothing_function=SmoothingFunction().method3,
            )
            bleu2 += sentence_bleu(
                label,
                pred,
                weights=[0.5, 0.5, 0, 0],
                smoothing_function=SmoothingFunction().method3,
            )
            bleu3 += sentence_bleu(
                label,
                pred,
                weights=[1 / 3, 1 / 3, 1 / 3, 0],
                smoothing_function=SmoothingFunction().method3,
            )
            bleu4 += sentence_bleu(
                label,
                pred,
                weights=[0.25, 0.25, 0.25, 0.25],
                smoothing_function=SmoothingFunction().method3,
            )
        bleu1 = round(bleu1 / len(ref_list)  * 100, 4)
        bleu2 = round(bleu2 / len(ref_list)  * 100, 4)
        bleu3 = round(bleu3 / len(ref_list)  * 100, 4)
        bleu4 = round(bleu4 / len(ref_list)  * 100, 4)
        mean_sent_bleu = round((bleu1 + bleu2 + bleu3 + bleu4) / 4, 4)
        
        eval_result.add(
            sent_bleu1=bleu1,
            sent_bleu2=bleu2,
            sent_bleu3=bleu3,
            sent_bleu4=bleu4,
            mean_sent_bleu=mean_sent_bleu,
        )
        
        return eval_result
    
    
class SacreSentenceBleuCritic:
    def __init__(self) -> None:
        self.name = "sacrebleu_sent"

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        bleu = []
        for i, a_gold in enumerate(references):
            bleu.append(sacrebleu.corpus_bleu([candidates[i]], [[a_gold]]).score)
        bleu_score = round(mean(bleu), 4)
        eval_result.add(sacrebleu_sent=bleu_score)
        return eval_result
    
    
class SacreCorpusBleuCritic:
    def __init__(self) -> None:
        self.name = "sacrebleu_word"

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        bleu = BLEU()
        bleu_score = round(bleu.corpus_score(candidates, [references]).score, 4)
        eval_result.add(sacrebleu_corpus=bleu_score)
        return eval_result
    

class MeteorCritic:
    def __init__(self) -> None:
        self.name = "meteor"

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        score_list = []
        for i in range(len(candidates)):
            if type(references[i]) is list:
                ref_list = references[i]
            else:
                ref_list = [references[i]]
            ref = [r.split(" ") for r in ref_list]
            cand = candidates[i].split(" ")
            score = meteor_scorer(ref, cand)
            score_list.append(score)
        meteor_score = round(mean(score_list), 4)
        eval_result.add(Meteor=meteor_score)
        return eval_result

class RougeCritic:
    def __init__(self) -> None:
        self.name = "rouge"

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        rouge = Rouge()
        scores = rouge.get_scores(candidates, references)
        rouge_1 = round(mean([score["rouge-1"]["f"] * 100 for score in scores]), 4)
        rouge_2 = round(mean([score["rouge-2"]["f"] * 100 for score in scores]), 4)
        rouge_l = round(mean([score["rouge-l"]["f"] * 100 for score in scores]), 4)
        
        eval_result.add(
            Rouge_1=rouge_1,
            Rouge_2=rouge_2,
            Rouge_l=rouge_l,
        )
        
        return eval_result


class BertScoreCritic:
    """  "en": "roberta-large",
         "zh": "bert-base-chinese",
    """
    
    def __init__(self) -> None:
        self.name = "bertscore"
        

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        scores = score(
            candidates,
            references,
            lang="en",
            verbose=False,
            rescale_with_baseline=True,
            device="cuda:0",
        )[-1].numpy()
        bert_score = round(mean(list(scores)), 4)
        eval_result.add(bert_score=bert_score)
        
        return eval_result

    
class ChrfCritic:
    def __init__(self) -> None:
        self.name = "chrf"

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        chrf = CHRF(word_order=2)
        chrf_score = round(chrf.corpus_score(candidates, [references]).score, 4)
        eval_result.add(
            chrf_score=chrf_score
        )
        return eval_result
    
    

    
class F1NLPCritic:
    def __init__(self) -> None:
        self.name = "f1"
        self.nlp = None
        
    def set_nlp(self):
        if self.nlp is None:
            self.nlp = spacy.load("en_core_web_sm")
        
    def get_tokens(self, text: str, nlp):
        doc = nlp(text)
        tokens = [tok.text.lower()
                for tok in doc if not tok.is_stop and not tok.is_punct]
        return tokens

    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        """
        This function is copied from: https://github.com/orhonovich/q-squared/blob/main/pipeline/score.py
        """
        self.set_nlp()
        f1_list = []
        for i, a_gold in enumerate(references):
            a_pred = candidates[i]
            if a_pred == "":
                f1_list.append(0)
                continue
            gold_toks = self.get_tokens(a_gold, self.nlp)
            pred_toks = self.get_tokens(a_pred, self.nlp)
            common = Counter(gold_toks) & Counter(pred_toks)
            num_same = sum(common.values())
            if num_same == 0:
                f1_list.append(0)
                continue
            precision = 1.0 * num_same / len(pred_toks)
            recall = 1.0 * num_same / len(gold_toks)
            f1 = (2 * precision * recall) / (precision + recall)
            f1_list.append(f1)
        f1 = round(sum(f1_list) / len(f1_list) * 100, 2)
        
        eval_result.add(
            F1_NLP_split=f1
        )
        return eval_result
    
    
    
    
class F1SpaceCritic:
    def __init__(self) -> None:
        self.name = "f1"
        
        
    def clean_text(self, text: str):
        text = text.lower()
        text = text.translate(str.maketrans("", "", string.punctuation))
        text = re.sub(r"\b(a|an|the|in|our)\b", " ", text)
        return re.sub(" +", " ", text).strip()


    def score(self, candidates: list[str], references: list[str], eval_result: Result, *args, **kwargs) -> Result:
        """
        This function is copied from: https://github.com/orhonovich/q-squared/blob/main/pipeline/score.py
        """
        self.set_nlp()
        f1_list = []
        for i, a_gold in enumerate(references):
            a_pred = candidates[i]
            if a_pred == "":
                f1_list.append(0)
                continue
            gold_toks = self.clean_text(a_gold).split()
            pred_toks = self.clean_text(a_pred).split()
            common = Counter(gold_toks) & Counter(pred_toks)
            num_same = sum(common.values())
            if num_same == 0:
                f1_list.append(0)
                continue
            precision = 1.0 * num_same / len(pred_toks)
            recall = 1.0 * num_same / len(gold_toks)
            f1 = (2 * precision * recall) / (precision + recall)
            f1_list.append(f1)
        f1 = round(sum(f1_list) / len(f1_list) * 100, 2)
        
        eval_result.add(
            F1_Space_split=f1
        )
        return eval_result
    
    
class Q2Critic:
    def __init__(self) -> None:
        self.name = "q2"
        
    def set_env_pkg(self):
        from q2.run_pipeline import q2_step_one
        from q2.run_nli import q2_step_two
        self.q2_step_one = q2_step_one
        self.q2_step_two = q2_step_two
        
    def score(self, candidates: list[str], references: list[str], eval_result: Result, device: int = 0, eval_dataset: Dataset = None, *args, **kwargs) -> Result:
        self.set_env_pkg()
        df, steps_df = self.q2_step_one(infile=pd.DataFrame(eval_dataset)[["model_responses", "references", "q2_references"]].rename(columns={
            "model_responses": "response",
            "references": "gold",
            "q2_references": "knowledge"
        }), gen_method="beam", q_per_cand="single", personal="remove", device=device)
        q_squared_nli, q_squared_f1 = self.q2_step_two(infile=steps_df, device=device)
        
        eval_result.add(
            q_squared_nli=round(q_squared_nli, 4),
            q_squared_f1=round(q_squared_f1, 4)
        )
        
        return eval_result
    
    


Metrics = {
    "corpus_bleu": CorpusBleuCritic(),
    "sent_bleu": SentenceBleuCritic(),
    "sent_sacrebleu": SacreSentenceBleuCritic(),
    "corpus_sacrebleu": SacreCorpusBleuCritic(),
    "meteor": MeteorCritic(),
    "rouge": RougeCritic(),
    "bert_score": BertScoreCritic(),
    "chrf": ChrfCritic(),
    "dist": DistCritic(),
    "f1_nlp": F1NLPCritic(),
    "f1_space": F1SpaceCritic(),
}


def get_eval_metrics(eval_inputs: Union[Dataset, pd.DataFrame, dict] = None, 
                     metrics: list = [], 
                     device: int = 0):
    """计算评价指标，目前已支持的指标有: corpus_bleu, sent_bleu, sent_sacrebleu, corpus_sacrebleu, meteor, rouge, bertscore, chrf, dist, f1_nlp, f1_space
    

    Args:
        eval_inputs (Union[Dataset, pd.DataFrame, dict], optional): _description_. Defaults to None.
        metrics (list, optional): _description_. Defaults to [].
        device (int, optional): _description_. Defaults to 0.

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    if not isinstance(eval_inputs, Dataset):
        if isinstance(eval_inputs, pd.DataFrame):
            eval_inputs = Dataset.from_pandas(eval_inputs)
        elif isinstance(eval_inputs, dict):
            eval_inputs = Dataset.from_dict(eval_inputs)
        else:
            raise ValueError("评价指标计算的输入 eval_inputs 必须是 Dataset, pd.DataFrame 或者 dict类型，并满足转为 Dataset 的条件。")
    eval_result = Result()
    if "model_responses" in eval_inputs.column_names:
        model_responses = eval_inputs["model_responses"]
    else:
        model_responses = eval_inputs["generated"]
    if "references" in eval_inputs.column_names:
        references = eval_inputs["references"]
    else:
        return eval_result
    
    if "" in model_responses:
        log.warning("评价指标计算的输入中（model_responses）包含空值，评测结果或有偏颇。")
        
    
    for metric in metrics:
        if metric not in Metrics:
            log.warning("评价指标 {} 不在可用的评价指标列表中，将跳过该指标。".format(metric))
            continue
        
        log.info("计算 {} 评价指标...".format(metric))
        
        try:
            eval_result = Metrics[metric].score(candidates=model_responses, references=references, eval_result=eval_result, device=device, eval_dataset=eval_inputs)
        except Exception as e:
            log.error(e)
            log.warning("评价指标 {} 计算失败，将跳过该指标。".format(metric))

    console = Console(color_system="256", style="cyan")
    table = Table(style="cyan", show_footer=False, title="[bold green]Evaluation results")
    table.add_column("Metric", justify="right", style="magenta")
    table.add_column("Score :yum:", justify="left", style="magenta")
    for k, v in eval_result.items():
        table.add_row(k, str(v))
    console.print(table)
    
    return eval_result
    







import transformers
from transformers import (
    PreTrainedModel,
    TrainingArguments,
    DataCollator,
    PreTrainedTokenizerBase,
    EvalPrediction,
    TrainerCallback,
)
from typing import Callable, Dict, List, Optional, Tuple, Union, Any
from torch import nn
from torch.utils.data import Dataset
from transformers.utils import (
    logging,
)
import os
import torch
import lightning.pytorch as L
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping, StochasticWeightAveraging
from lightning.pytorch.loggers import CometLogger
from lightning.pytorch.callbacks.progress import TQDMProgressBar
from lightning.pytorch.utilities import rank_zero_only
from colorama import Fore



logger = logging.get_logger(__name__)

# Name of the files used for checkpointing
TRAINING_ARGS_NAME = "training_args.bin"
TRAINER_STATE_NAME = "trainer_state.json"
OPTIMIZER_NAME = "optimizer.pt"
SCHEDULER_NAME = "scheduler.pt"
SCALER_NAME = "scaler.pt"


class HFTrainer(transformers.Trainer):
    """
    主要修改逻辑：通过传入compute_loss，支持自定义loss计算方式
    """
    def __init__(
            self,
            model: Union[PreTrainedModel, nn.Module] = None,
            args: TrainingArguments = None,
            data_collator: Optional[DataCollator] = None,
            train_dataset: Optional[Dataset] = None,
            eval_dataset: Optional[Dataset] = None,
            tokenizer: Optional[PreTrainedTokenizerBase] = None,
            model_init: Callable[[], PreTrainedModel] = None,
            compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
            callbacks: Optional[List[TrainerCallback]] = None,
            optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR] = (None, None),
            preprocess_logits_for_metrics: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
            compute_loss=None,
    ):
        super(Trainer_HF, self).__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            model_init=model_init,
            compute_metrics=compute_metrics,
            callbacks=callbacks,
            optimizers=optimizers,
            preprocess_logits_for_metrics=preprocess_logits_for_metrics,
        )
        self.loss_func = compute_loss

    def compute_loss(self, model, inputs, return_outputs=False):
        """
        重写loss的计算方式
        How the loss is computed by Trainer. By default, all models return the loss in the first element.

        Subclass and override for custom behavior.
        """
        return self.loss_func(model, inputs, self.args, return_outputs)
    
    """
    修改checkkpoint的保存逻辑，只保存lora
    """
    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        # If we are executing this function, we are the process zero, so we don't check for that.
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info(f"Saving model checkpoint to {output_dir}")
        # 保存lora权重和配置
        self.model.save_pretrained(
            output_dir, state_dict=state_dict, safe_serialization=self.args.save_safetensors
        )

        if self.tokenizer is not None:
            self.tokenizer.save_pretrained(output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(self.args, os.path.join(output_dir, TRAINING_ARGS_NAME))



class LitTrainer:
    def __init__(
        self,
        config,
        model,
        train_dataloader,
        val_dataloader,
        tokenizer,
        experiment,
    ):
        self.config = config
        self.model = model
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.tokenizer = tokenizer
        
        
        if experiment:
            logger = CustomCometLoggerForLit()
            logger._experiment = experiment
        else:
            logger = None

        # checkpoint_callback = ModelCheckpoint(
        #     monitor="val_loss",
        #     mode="min",
        #     save_top_k=config.save_top_k,
        #     save_last=True,
        #     verbose=True,
        #     dirpath=config.output_dir,
        #     filename="best_model",
        #     auto_insert_metric_name=False,
        # )

        early_stop_callback = EarlyStopping(
            monitor="val_loss",
            min_delta=0.001,
            patience=5,
            verbose=True,
            mode="min",
        )

        callbacks = [
            early_stop_callback,
            # checkpoint_callback,
            LiteProgressBar(),
        ]

        if config.get("use_swa"):
            callbacks.append(StochasticWeightAveraging(swa_lrs=1e-2))

        self.trainer = L.Trainer(
            logger=logger, callbacks=callbacks, **config.lit_args
        )


    def train(self):
        if self.config.auto_lr_find:
            lr_finder = self.trainer.tuner.lr_find(self.model, train_dataloaders=self.train_dataloader)
            # 展示loss和学习率的曲线
            fig = lr_finder.plot(suggest=True)
            fig.show()
            # 设置为推荐的学习率
            self.model.config.lr = lr_finder.suggestion()
            
        return self.trainer.fit(model=self.model,
                                train_dataloaders=self.train_dataloader,
                                val_dataloaders=self.val_dataloader)
        
    def save_model(self, save_path: str):
        self.model.backbone.save_pretrained(save_path)
        self.tokenizer.save_pretrained(save_path)
        logger.info(f"Model and tokenizer have saved to {save_path}")
        
        

class CustomCometLoggerForLit(CometLogger):
    def __init__(self):
        super(CustomCometLoggerForLit, self).__init__()

    @rank_zero_only
    def finalize(self, status: str) -> None:
        r"""
        When calling ``self.experiment.end()``, that experiment won't log any more data to Comet.
        That's why, if you need to log any more data, you need to create an ExistingCometExperiment.
        For example, to log data when testing your model after training, because when training is
        finalized :meth:`CometLogger.finalize` is called.

        This happens automatically in the :meth:`~CometLogger.experiment` property, when
        ``self._experiment`` is set to ``None``, i.e. ``self.reset_experiment()``.
        """
        # self.experiment.end()
        # self.reset_experiment()
        
        
        
class LiteProgressBar(TQDMProgressBar):
    def __init__(self, refresh_rate: int = 1, process_position: int = 0):
        super().__init__(refresh_rate, process_position)

    def get_metrics(
        self, trainer: "L.Trainer", pl_module: "L.LightningModule"
    ) -> Dict[str, Union[int, str]]:
        items = super().get_metrics(trainer, pl_module)
        # items['ppl'] = round(items['ppl'], 1) if 'ppl' in items else None
        items["lr"] = round(items["lr"], 7) if "lr" in items else None
        items.pop("v_num", None)
        return items

    def init_train_tqdm(self):
        bar = super().init_train_tqdm()
        bar.bar_format = "%s{l_bar}%s{bar}%s{r_bar}" % (
            Fore.GREEN,
            Fore.GREEN,
            Fore.GREEN,
        )
        return bar

    def init_validation_tqdm(self):
        bar = super().init_validation_tqdm()
        bar.set_description("Validating")
        bar.bar_format = "%s{l_bar}%s{bar}%s{r_bar}" % (
            Fore.GREEN,
            Fore.GREEN,
            Fore.GREEN,
        )
        bar.leave = False
        return bar


from datasets import load_dataset


def load_hf_faithdial(*args, **kwargs):
    """加载 FaithDial 数据集（英文），对话领域，对 WoW 数据集中的幻觉进行了修正
    
    train: 36809 turns
    valid: 6851 turns
    test: 7101 turns
    
    [
        {
            "utterances": [
            ... // prior utterances, 
            {
                "history": [
                "Have you ever been to a concert? They're so fun!",
                ],
                "speaker": "Wizard",
                "knowledge": "It began on September 9, 2015, in Montreal, Canada, at the Bell Centre and concluded on March 20, 2016, in Sydney, Australia at Allphones Arena.",
                "original_response": "It started in September of 2015 and ran all the way through March of 2016. Can you imagine being on the road that long?",
                "response": "Sure. The concert started in September 9th of 2015 at Montreal, Canada. It continued till 20th of March of 2016, where it ended at Sydney, Australia.",
                "BEGIN": [
                "Hallucination",
                "Entailment"
                ],
                "VRM": [
                "Disclosure",
                "Question"
                ]
            }, 
            ... // more utterances
            ]
        }, 
        ... // more dialogues
    ]
    
    """
    return load_dataset("McGill-NLP/FaithDial", *args, **kwargs)


def load_hf_wizard_of_wikipedia(*args, **kwargs):
    """加载 WoW 数据集（英文），对话领域, 一个对话者随机选择一个初始话题，对话双方可以在此基础上进行对话，但在对话过程中话题也可以拓展。对话双方的角色是不同的，分为 wizard 和 apprentice：

        wizard：wizard 的目的是通知 apprentice 关于对话主题相关的背景知识，在对话开始之前，会给定一些相关的 wiki 段落，这些对于 apprentice 不可见。同时，wizard 不允许直接复制拷贝 wiki 里的文本句子作为回复，而是需要自己进行组合生成融合知识的回答。
        apprentice：apprentice 的目的是深入的询问与对话主题相关的问题，这与普通的闲聊有所区别。
    
    测试集分为 seen 和 unseen 两种，seen 测试集中的话题在训练集中出现过，unseen 测试集中的话题在训练集中没有出现过。
    
    整个数据集超过 2W 个对话，540W 篇文档

    """
    return load_dataset("chujiezheng/wizard_of_wikipedia", *args, **kwargs)


def load_hf_CEval(*args, **kargs):
    """加载 C-Eval 数据集（中文），C-Eval是全面的中文基础模型评估套件，涵盖了52个不同学科的13948个多项选择题，分为四个难度级别
    
    Github: https://github.com/SJTU-LIT/ceval/tree/main

    Returns:
        _type_: _description_
    """
    return load_dataset("ceval/ceval-exam", *args, **kargs)


def load_AGIEval(dir: str = None, prompt_path: str = None, *args, **kargs):
    """加载 A-GIEval 数据集（中文），A-GIEval是全面的中文基础模型评估套件，涵盖了52个不同学科的13948个多项选择题，分为四个难度级别
    
    Github: https://github.com/microsoft/AGIEval
    prompt_data_csv: https://github.com/microsoft/AGIEval/blob/main/data/few_shot_prompts.csv
    
    {
        "passage": null,
        "question": "设集合 $A=\\{x \\mid x \\geq 1\\}, B=\\{x \\mid-1<x<2\\}$, 则 $A \\cap B=$ ($\\quad$)\\\\\n",
        "options": ["(A)$\\{x \\mid x>-1\\}$", 
            "(B)$\\{x \\mid x \\geq 1\\}$", 
            "(C)$\\{x \\mid-1<x<1\\}$", 
            "(D)$\\{x \\mid 1 \\leq x<2\\}$"
            ],
        "label": "D",
        "answer": null
    }
    
    """
    print("AGIEval 数据集正在开发中...")
    from lazydl.utils.data_utils.AGIEval import post_process, utils, dataset_loader
    dataset_name_list = [
        "aqua-rat",
        "math",
        "logiqa-en", "logiqa-zh",
         "jec-qa-kd", "jec-qa-ca",
        "lsat-ar", "lsat-lr", "lsat-rc",
        "sat-math", "sat-en",
        "sat-en-without-passage",
        "gaokao-chinese",
        "gaokao-english",
        "gaokao-geography", "gaokao-history",
        "gaokao-biology", "gaokao-chemistry", "gaokao-physics",
        "gaokao-mathqa",
        "gaokao-mathcloze",
    ]
    setting_name_list = [
        'zero-shot',
        'zero-shot-CoT',
        'few-shot',
        'few-shot-CoT',
    ]
    datasets = {}
    for dataset_name in dataset_name_list:
        for setting_name in setting_name_list:
            if 'few-shot' in setting_name:
                continue
            dataset = dataset_loader.load_dataset(
                dataset_name, setting_name, dir,
                prompt_path=prompt_path, max_tokens=2048,
                end_of_example="<END>\n", verbose=True)
            datasets[f"{dataset_name}-{setting_name}"] = dataset
            
    return datasets
    


import comet_ml
from omegaconf import DictConfig, OmegaConf
import os
import json
import sys
from lazydl.utils.log import Logger
from typing import Union
import datetime
import numpy

log = Logger(__name__) 


def init_env(config: Union[DictConfig, dict], current_dir: str = "./") -> None:
    """初始化环境，包括随机种子、可见GPU、Comet.ml环境变量、进程名

    Args:
        config (DictConfig, dict): 配置文件
            seed (str, optional): 种子. Defaults to '3407'.
            use_deterministic (bool, optional): 是否使用确定性算法，使用了训练会变慢. Defaults to False.
            visibale_cuda (str, optional): 设置可见 GPU，如果不使用设置为 ''. Defaults to 'all'.
            proctitle (str, optional): 进程名. Defaults to 'python'.
            proctitle_prefix_id (bool, optional): 是否在进程名前边添加进程 ID. Defaults to True.
    """
    if isinstance(config, dict):
        config = OmegaConf.create(config)
    OmegaConf.set_struct(config, False)
    updated_env = {}
    
    config.root_dir = current_dir
    if config.get("stage", "debug") == "debug":
        log.info("Debug mode will set fast_dev_run to True")
        config.fast_dev_run = True
    
    # ---------------------------------------------------------------------------- #
    #                         生成 Task ID                                     
    # ---------------------------------------------------------------------------- #
    task_id_prefix = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    config.task_id = "{}{}{}".format(task_id_prefix, ''.join([str(num) for num in numpy.random.randint(0, 9, size=18)]), os.getpid())
    
    
    # ---------------------------------------------------------------------------- #
    #                         设置可见 GPU                                                          
    # ---------------------------------------------------------------------------- #
    if config.get("env.visibale_cuda", "all") != "all":
        os.environ['CUDA_VISIBLE_DEVICES'] = config.get("visibale_cuda")
        updated_env['CUDA_VISIBLE_DEVICES'] = os.environ['CUDA_VISIBLE_DEVICES']
    
    # ---------------------------------------------------------------------------- #
    #                         设置随机种子                                           
    # ---------------------------------------------------------------------------- #
    seed = config.get("seed", 3407)
    use_deterministic = config.get("use_deterministic", False)
    import random
    import numpy as np
    import torch
    random.seed(seed)   # Python的随机性
    os.environ['PYTHONHASHSEED'] = str(seed)    # 设置Python哈希种子，为了禁止hash随机化，使得实验可复现
    updated_env['PYTHONHASHSEED'] = os.environ['PYTHONHASHSEED']
    np.random.seed(seed)   # numpy的随机性
    torch.manual_seed(seed)   # torch的CPU随机性，为CPU设置随机种子
    torch.cuda.manual_seed(seed)   # torch的GPU随机性，为当前GPU设置随机种子
    torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.   torch的GPU随机性，为所有GPU设置随机种子
    torch.backends.cudnn.benchmark = not use_deterministic # if benchmark=True, deterministic will be False
    torch.backends.cudnn.deterministic = use_deterministic   # 选择确定性算法
    
    # ---------------------------------------------------------------------------- #
    #                         设置 Comet.ml 环境变量                                     
    # ---------------------------------------------------------------------------- #
    os.environ["COMET_API_KEY"] = "" if config.get("comet_api_key") is None else config.get("comet_api_key")
    
    updated_env['COMET_API_KEY'] = os.environ['COMET_API_KEY']
    
    # ---------------------------------------------------------------------------- #
    #                         设置 Comet Project                                     
    # ---------------------------------------------------------------------------- #
    experiment = None
    if config.get("start_comet_log", "False"):
        api = comet_ml.api.API()
        experiment = api.get_experiment_by_key(config.get("task_id"))
        
        if experiment is None:
            experiment = comet_ml.Experiment(project_name=config.get("comet_project_name", "no-comet-project"),
                                            experiment_key=config.get("task_id"),)
            experiment.set_name(config.get("comet_exp_name", "NULL") + "_" + config.get("stage", "debug"))
            
            experiment_config = sys.argv[-1].replace("+experiments=", "")

        
        tmux_session = "/"
        for arg in sys.argv:
            if "tmux_session" in arg:
                tmux_session = arg.replace("+tmux_session=", "")
                
        experiment.log_other("tmux_session", tmux_session)
        experiment.log_other("进程ID", str(os.getpid()))
        experiment.log_other("实验配置文件", experiment_config)
        for key, value in config.items():
            if isinstance(value, dict):
                value = json.dumps(value)
            experiment.log_other(key, str(value))
        for tag in config.get("comet_tags", []):
            experiment.add_tag(tag)
        log.info("Comet 实验记录已启动")
    
    # ---------------------------------------------------------------------------- #
    #                         设置钉钉 Token                                     
    # ---------------------------------------------------------------------------- #
    os.environ["DINGDING_ACCESS_TOKEN"] = config.get("dingding_access_token") if config.get("dingding_access_token") else "NONE_DINGDING_ACCESS_TOKEN"
    os.environ["DINGDING_SECRET"] = config.get("dingding_secret") if config.get("dingding_secret") else "NONE_DINGDING_SECRET"
    
    updated_env['DINGDING_ACCESS_TOKEN'] = os.environ['DINGDING_ACCESS_TOKEN']
    updated_env['DINGDING_SECRET'] = os.environ['DINGDING_SECRET']
    
        
    # ---------------------------------------------------------------------------- #
    #                         设置其他环境变量                                     
    # ---------------------------------------------------------------------------- #
    os.environ["TOKENIZERS_PARALLELISM"] = config.get("tokenizers_parallelism", "False")
    updated_env['TOKENIZERS_PARALLELISM'] = os.environ['TOKENIZERS_PARALLELISM']
    
    
    # ---------------------------------------------------------------------------- #
    #                         设置进程名                                     
    # ---------------------------------------------------------------------------- #
    import setproctitle
    if config.get("proctitle_prefix_id", True):
        setproctitle.setproctitle(str(os.getpid()) + "| " + config.get("proctitle", "python"))
    else:
        setproctitle.setproctitle(config.get("proctitle", "python"))
        
    # ---------------------------------------------------------------------------- #
    #                        打印环境变量                                      
    # ---------------------------------------------------------------------------- #
    log.info(json.dumps(updated_env, indent=4, ensure_ascii=False).replace("{", "Environment Variables\n").replace("}", ""))
    
    # ---------------------------------------------------------------------------- #
    #                         设置处理器类型                                     
    # ---------------------------------------------------------------------------- #
    processing_unit = config.get("processing_unit", "cpu")
    if processing_unit == "cpu":
        config.processing_unit_type = "cpu"
    elif processing_unit == "mps":
        config.processing_unit_type = "mps"
    else:
        if not torch.cuda.is_available():
            log.error("CUDA is not available, please check your CUDA installation.")
            raise Exception("CUDA is not available, please check your CUDA installation.")
        if "^" in processing_unit:
            # 自动选择 GPU
            processing_unit = int(processing_unit.replace("^", ""))
            if processing_unit > torch.cuda.device_count():
                log.error("The number of GPUs you specified is greater than the number of GPUs available.")
                raise Exception("The number of GPUs you specified is greater than the number of GPUs available.")
            if processing_unit > 1:
                config.processing_unit_type = "gpu-a-m"
            else:
                config.processing_unit_type = "gpu-a-s"
        else:
            processing_unit = processing_unit.split(",")
            if len(processing_unit) > torch.cuda.device_count():
                log.error("The number of GPUs you specified is greater than the number of GPUs available.")
                raise Exception("The number of GPUs you specified is greater than the number of GPUs available.")
            if len(processing_unit) > 1:
                config.processing_unit_type = "gpu-m-m"
            else:
                config.processing_unit_type = "gpu-m-s"
        config.processing_unit = processing_unit
        
    log.info(f"Processing unit type: {config.get('processing_unit_type')}")
    log.info(f"Processing unit num: {config.get('processing_unit')}\n")
    
    return config, experiment




import logging
import sys
import os
import inspect
import time


class Logger:
    COLORS = {
        'DEBUG': '\033[94m',
        'INFO': '\033[92m',
        'WARNING': '\033[93m',
        'ERROR': '\033[31m',
        'CRITICAL': '\033[91m',
        'FILENAME': '\033[95m',
        'LINENO': '\033[95m',
    }
    RESET = '\033[0m'

    def __init__(self, name):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)
        self.rank = 0

    def get_logger(self):
        return self.logger

    def set_rank(self, rank):
        self.rank = rank

    def format_with_color(self, message, color):
        return f'{color}{message}{self.RESET}'

    def debug(self, message):
        if self.rank == 0:
            frame = inspect.currentframe().f_back
            filename = os.path.basename(frame.f_code.co_filename)
            lineno = frame.f_lineno
            timestamp = time.strftime('%m-%d %H:%M:%S', time.localtime())
            formatted_time = self.format_with_color(timestamp, self.COLORS['DEBUG'])
            formatted_filename = self.format_with_color(filename, self.COLORS['FILENAME'])
            formatted_lineno = self.format_with_color(str(lineno), self.COLORS['LINENO'])
            print(f'\033[1m{self.COLORS["DEBUG"]}[DEBUG] {formatted_time}  {self.RESET} \033[1m[{formatted_filename}:\033[1m{formatted_lineno}] \033[1m{message}')

    def info(self, message):
        if self.rank == 0:
            frame = inspect.currentframe().f_back
            filename = os.path.basename(frame.f_code.co_filename)
            lineno = frame.f_lineno
            timestamp = time.strftime('%m-%d %H:%M:%S', time.localtime())
            formatted_time = self.format_with_color(timestamp, self.COLORS['INFO'])
            formatted_filename = self.format_with_color(filename, self.COLORS['FILENAME'])
            formatted_lineno = self.format_with_color(str(lineno), self.COLORS['LINENO'])
            print(f'\033[2m{formatted_time} {self.COLORS["INFO"]}{self.RESET} \033[2m[{formatted_filename}:\033[2m{formatted_lineno}] {message}')

    def warning(self, message):
        if self.rank == 0:
            frame = inspect.currentframe().f_back
            filename = os.path.basename(frame.f_code.co_filename)
            lineno = frame.f_lineno
            timestamp = time.strftime('%m-%d %H:%M:%S', time.localtime())
            formatted_time = self.format_with_color(timestamp, self.COLORS['WARNING'])
            formatted_filename = self.format_with_color(filename, self.COLORS['FILENAME'])
            formatted_lineno = self.format_with_color(str(lineno), self.COLORS['LINENO'])
            print(f'\033[1m{self.COLORS["WARNING"]}[WARNING] {formatted_time} {self.RESET} \033[1m[{formatted_filename}:\033[1m{formatted_lineno}] \033[1m{message}')

    def error(self, message):
        if self.rank == 0:
            frame = inspect.currentframe().f_back
            filename = os.path.basename(frame.f_code.co_filename)
            lineno = frame.f_lineno
            timestamp = time.strftime('%m-%d %H:%M:%S', time.localtime())
            formatted_time = self.format_with_color(timestamp, self.COLORS['ERROR'])
            formatted_filename = self.format_with_color(filename, self.COLORS['FILENAME'])
            formatted_lineno = self.format_with_color(str(lineno), self.COLORS['LINENO'])
            print(f'\033[1m{self.COLORS["ERROR"]}[ERROR] {formatted_time} {self.RESET} \033[1m[{formatted_filename}:\033[1m{formatted_lineno}] \033[1m{message}')

    def critical(self, message):
        if self.rank == 0:
            frame = inspect.currentframe().f_back
            filename = os.path.basename(frame.f_code.co_filename)
            lineno = frame.f_lineno
            timestamp = time.strftime('%m-%d %H:%M:%S', time.localtime())
            formatted_time = self.format_with_color(timestamp, self.COLORS['CRITICAL'])
            formatted_filename = self.format_with_color(filename, self.COLORS['FILENAME'])
            formatted_lineno = self.format_with_color(str(lineno), self.COLORS['LINENO'])
            print(f'\033[1m{self.COLORS["CRITICAL"]}[CRITICAL] {formatted_time} {self.RESET} \033[1m[{formatted_filename}:\033[1m{formatted_lineno}] \033[1m{message}')

